// Component TBASSPlayer Ver 1.3
//
//    written by Silhwan Hyun  hsh@chollian.net
//
//
// Revision history
//
// Ver 1.3                        5 Apr 2003
//   ( Based on bass.dll/basswma.dll Ver 1.8 )
//  Added a function and a procedure to run Winamp Visualization plug-ins
//   - function RunVisoutProgram(VismodName : string; VismodNo : word) : boolean;
//   - procedure QuitVisoutProgram;
//  Modified function GetVolume and procedure SetVolume to reflect/adjust the
//   volume level of Winamp plug-in's output
//  Added property EchoLevel & ReverbLevel to adjust the intensity of sound effect
//
// Ver 1.2                        5 Feb 2003
//   ( Based on bass.dll/basswma.dll Ver 1.7 )
//  Added functions & properties to use Winamp 2 input plug-ins.
//   Now TBASSPlayer can play stream files which BASS sound system cannot decode
//   if provided appropriate Winamp input plug-in.
//  Changed the loading method of bass.dll & basswma.dll from static loading to
//   dynamic loading
//
//   - Modified function : GetStreamInfo
//   - New properties : BASSWMAReady, Version, NativeFileExts, PluginFileExts, CanUsePlugin,
//                      SupportedBy, PluginFirst, OnPluginStartPlay, UnloadButEnabled
//   - Renamed property : DLLVersionStr -> BASSDLLVer
//   - New  procedure : ShowPluginConfigForm (-> used to show PluginConfigForm,
//                        which is defined in sub unit PluginConfig.pas)
//   - New functions in sub unit PluginCtrl which you can call in main program
//     if you want to manage Winamp input plug-ins directly(= not using PluginConfigForm).
//       GetPluginIndex(PluginName : string) : integer;
//       GetPlugin(PluginNum : integer) : TPlugin;
//       GetPluginInfo(PluginNum : integer) : TPluginInfo;
//       LoadWinampPlugin(PluginName : string) : integer;
//       UnloadWinampPlugin(PluginName : string) : integer;
//  * include "PluginCtrl" in uses clause of main program to use above functions
//
//  * The items which does not affect external program are not mentioned here
//
// Ver 1.1                         8 Dec 2002
//   ( Based on bass.dll/basswma.dll Ver 1.7 )
//   - New function : GetChannelFFTData (for custom processing of FFT data)
//   - Added field at record TStreamInfo : Year
//   - Added property : EQBands (for custom setting of equalizer bands)
//   - Renamed property : EQGains ( <- EqPreset )
//   - New type identifiers : TBandData, TEQBands
//   - Renamed type identifier : TEQGains ( <- TEqPreset )
//   - Updated procedure : TimerFFTTimer (for better output of spectrum visualization)
//
// Ver 1.0                         4 Aug 2002
//   ( Based on bass.dll/basswma.dll Ver 1.6 )
//   - Initial release

unit BASSPlayer;

interface

uses
  Windows, Messages, SysUtils, Classes, Forms, Controls, StdCtrls, ExtCtrls,
  RT_BASS, RT_BASSWMA, ioplug, PluginCtrl, MPEGAudio, OggVorbis, WMAFile, WAVFile,
  PluginConfig, appexec, fisSharedMemory, Dialogs;

const
   Ver = '1.2';
   NumFFTBands = 25;     // Number of bands for spectrum visualization
   NumEQBands  = 10;     // Number of bands for equalizer
   EQBandWidth  : array[0..NumEQBands-1] of float =
                  (4, 4, 4, 4, 5, 6, 5, 4, 3, 3);
   EQFreq : array[0..NumEQBands-1] of float =
                  (80, 170, 310, 600, 1000, 3000, 6000, 10000, 12000, 14000);
   NoDX8Msg : pchar = 'DirectX Ver 8 are not installed.' + #13#10 +
                      'Sound effects except flanger are disabled.';

   WM_GetToEnd      = WM_USER + 107;
   WM_NewFFTData    = WM_USER + 108;

type
  TSoundEffects = set of (Flanger, Equalizer, Echo, Reverb);
  TStreamInfo = record
     FileName : string;
     FileSize : DWORD;        // File size in byte
     SampleRate : LongInt;    // Sampling rate in Hz
     BitRate : LongInt;       // Bit Rate in KBPS
     Duration : DWORD;        // Song duration in mili second
     Channels : Word;         // 1- Mono, 2 - Stereo

     Title : string;
     Artist : string;
     Album : string;
     Year : string;
     Genre : string;
     GenreID : byte;
     Track : byte;
     Comment : string;
  end;

  TFFTData = array [0..512] of Single;
  TBandOut = array[0..NumFFTBands-1] of word;
  TPlayerMode = (plmStandby, plmReady, plmStopped, plmPlaying, plmPaused);
  TChannelType = (Channel_NotOpened, Channel_Stream, Channel_WMA, Channel_Music, Channel_Plugin);
  TSupportedBy = (Both, BASSNative, WinampPlugin, None);
  TEQGains = array [0..NumEQBands-1] of float;
  TBandData = record
     CenterFreq : float;
     Bandwidth  : float;
  end;
  TEQBands = record
     Bands : word;     // Number of equalizer bands (min : 0, max : 10)
     BandPara : array[0..NumEQBands-1] of TBandData;
  end;

  TNotifyEvent = procedure(Sender: TObject) of object;
  TNotifyPluginEvent = procedure(Sender: TObject; Channels : WORD; SampleRate : Cardinal) of object;
  TNotifyNewFFTDataEvent = procedure(Sender: TObject; BandOut : TBandOut) of object;

  TBASSPlayer = class(TComponent)
  private
    { Private declarations }
    AppHWND     : THandle;
    TimerFFT    : TTimer;

    ChannelType   : TChannelType;
    Channel       : DWORD;
    MessageHandle : hwnd;

    APlugin : TPlugin;
    MessageHandle2 : hwnd;
    NowStarting : boolean;
    UsesOutputPlugin : boolean;
 // Some input plug-ins do not need output plug-in. i.e., some input plug-ins
 // sound out decoded data directly.
 // UsesOutputPlugin is used to know if input plug-in needs output plug-in.
 // Use this instead of input plug-in's UsesOutPlug property because some
 // input plug-ins report wrong UsesOutPlug value.

    BASSDLLLoaded : boolean;
    BassInfoParam : BASS_INFO;
    EQParam       : BASS_FXPARAMEQ;
    EchoParam     : BASS_FXECHO;
    ReverbParam   : BASS_FXREVERB;
    FSoundEffects : TSoundEffects;
    SyncHandle    : HSYNC;
    fladspHandle  : HDSP;
    EQHandle      : array [0..NumEQBands-1] of HFX;
    EchoHandle    : HFX;
    ReverbHandle  : HFX;

    BandOut: TBandOut;
    fisSharedMemory1: TfisSharedMemory;
    ShareMemPointer : pointer;

    MPEG        : TMPEGaudio;
    Vorbis      : TOggVorbis;
    WMA         : TWMAfile;
    WAV         : TWAVFile;

    AppExec1: TAppExec;
    GoVisOut : boolean;
    
    PluginConfigForm: TPluginConfigForm;

    FBASSReady      : boolean;
    FBASSWMAReady   : boolean;
    FPluginNum      : integer;
    FPluginFirst    : boolean;
    FVersionStr     : string;
    FDLLVersionStr  : string;
    FStreamName     : string;
    FSupportedBy    : TSupportedBy;
    FStreamInfo     : TStreamInfo;
    FMusicVolume    : integer;
    FStreamVolume   : integer;
    FPluginVolume   : integer;      // * New at Ver 1.3
    FDX8EffectReady : boolean;
    FEchoLevel      : word;
    FReverbLevel    : word;
    FEQGains        : TEQGains;
    FEQBands        : TEQBands;
    FPlayerMode     : TPlayerMode;
    FUnloadButEnabled : boolean;

    FOnPlayEnd         : TNotifyEvent;
    FOnPluginStartPlay : TNotifyPluginEvent;
    FOnNewFFTData      : TNotifyNewFFTDataEvent;

    procedure ClearEffectHandle;                   // * New at Ver 1.2
    function  IsPluginReady : boolean;             // * New at Ver 1.2
    function  GetStreamInfo2(StreamName : string;  // * New at Ver 1.2
                            var StreamInfo : TStreamInfo;
                            var SupportedBy : TSupportedBy;
                            var PluginNum : integer) : boolean;
    function  GetVolume : integer;            // * Modified at Ver 1.3
    procedure SetVersionStr(Value : string);
    procedure SetDLLVersionStr(Value : string);
    procedure SetDX8EffectReady(Value : boolean);
    procedure SetVolume(Value : integer);     // * Modified at Ver 1.3
    procedure SetEchoLevel(Value : word);     // * New at Ver 1.3
    procedure SetReverbLevel(Value : word);   // * New at Ver 1.3
    procedure SetSoundEffect(Value : TSoundEffects);
    procedure SetEQGains(Value : TEQGains);
    procedure SetEQBands(Value : TEQBands);
    procedure SetPluginFirst(Value : boolean); // * New at Ver 1.2
    procedure SetUnloadBut(Value : boolean);   // * New at Ver 1.2
    function  GetDLLVersionStr : string;
    function  GetNativeFileExts : string;      // * New at Ver 1.2
    function  GetPluginFileExts : string;      // * New at Ver 1.2
    procedure PausePlay;
    procedure ResumePlay;
    procedure Restart;
    procedure TimerFFTTimer(Sender: TObject);
    procedure ProcMessage(var Msg: TMessage);
    function  CurrentPosition : DWORD;
    procedure SetPosition(Value : DWORD);
    procedure Error(msg: string);
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;

    function  GetStreamInfo(StreamName : string;        // * Modified at Ver 1.2
                            var StreamInfo : TStreamInfo;
                            var SupportedBy : TSupportedBy) : boolean;
    function  Open(StreamName : string) : boolean;

    function  PlayLength : DWORD;
    function  GetChannelFFTData(PFFTData : pointer; FFTFlag : DWORD) : boolean;

    procedure Play;
    procedure Stop;
    procedure Pause(pAction : boolean);
    procedure Close;

    procedure ShowPluginConfigForm;                          // * New at Ver 1.2
    property  EQGains : TEQGains read FEQGains write SetEQGains;
    property  EQBands : TEQBands read FEQBands write SetEQBands;
    property  PlayerReady : boolean read FBASSReady;
    property  BASSWMAReady : boolean read FBASSWMAReady;     // * New at Ver 1.2
    property  SupportedBy : TSupportedBy read FSupportedBy;  // * New at Ver 1.2
    property  StreamInfo : TStreamInfo read FStreamInfo;
    property  Mode : TPlayerMode read FPlayerMode;
    property  Position : DWORD read CurrentPosition write SetPosition;
    property  StreamName : string read FStreamName;

    property  Volume : integer read GetVolume write SetVolume;  // 0(minimum output) ~ 100(maximum output)
    property  SoundEffects : TSoundEffects read FSoundEffects write SetSoundEffect;

  // EchoLevel & ReverbLevel : 0(minimum sound effect) ~ 32(maximum sound effect)
    property  EchoLevel : word read FEchoLevel write SetEchoLevel;        // * New at Ver 1.3
    property  ReverbLevel : word read FReverbLevel write SetReverbLevel;  // * New at Ver 1.3

    property  CanUsePlugin : boolean read IsPluginReady;      // * New at Ver 1.2
    property  NativeFileExts : string read GetNativeFileExts; // * New at Ver 1.2
    property  PluginFileExts : string read GetPluginFileExts; // * New at Ver 1.2
    property  UnloadButEnabled : boolean read FUnloadButEnabled write SetUnloadBut; // * New at Ver 1.2

    function  RunVisoutProgram(VismodName : string; VismodNo : word) : boolean; // * New at Ver 1.3
    procedure QuitVisoutProgram;       // * New at Ver 1.3
  published
    { Published declarations }
    property  Version : string read FVersionStr write SetVersionStr;    // * New at Ver 1.2
    property  BASSDLLVer : string read FDLLVersionStr write SetDLLVersionStr; // * Renamed at Ver 1.2
    property  DX8EffectReady : boolean read FDX8EffectReady write SetDX8EffectReady;
    property  PluginFirst : boolean read FPluginFirst write SetPluginFirst;  // * New at Ver 1.2
    property  OnNewFFTData : TNotifyNewFFTDataEvent read FOnNewFFTData write FOnNewFFTData;
    property  OnPluginStartPlay : TNotifyPluginEvent read FOnPluginStartPlay
                                           write FOnPluginStartPlay; // * New at Ver 1.2
    property  OnPlayEnd : TNotifyEvent read FOnPlayEnd write FOnPlayEnd;
  end;

  procedure Register;


implementation


procedure WinProcessMessages;
// Allow Windows to process other system messages
var
    ProcMsg  :  TMsg;
begin
    while PeekMessage(ProcMsg, 0, 0, 0, PM_REMOVE) do begin
      if (ProcMsg.Message = WM_QUIT) then Exit;
      TranslateMessage(ProcMsg);
      DispatchMessage(ProcMsg);
    end;
end;


// Following statements for flanger effect is adopted from DTMain.pas of BASS16
// package

const
   FLABUFLEN = 350;         // buffer length for flanger effect

var
   // Variables for DSP (flanger effect) implementation
   flabuf : array[0..FLABUFLEN-1, 0..2] of SmallInt;  // buffer
   flapos : Integer;         // cur.pos
   flas, flasinc : FLOAT;    // sweep pos/min/max/inc

function fmod(a, b: FLOAT): FLOAT;
begin
   Result := a - (b * Trunc(a / b));
end;

function Clip(a: Integer): Integer;
begin
   if a <= -32768 then
      a := -32768
   else if a >= 32767 then
      a := 32767;

   Result := a;
end;

procedure Flange(handle: HSYNC; channel: DWORD; buffer: Pointer; length, user: DWORD); stdcall;
var
  lc, rc: SmallInt;
  p1, p2, s: Integer;
  d: ^DWORD;
  f: FLOAT;
begin
  d := buffer;
  while (length > 0) do
  begin
    lc := LOWORD(d^); rc := HIWORD(d^);
    p1 := (flapos + Trunc(flas)) mod FLABUFLEN;
    p2 := (p1 + 1) mod FLABUFLEN;
    f := fmod(flas, 1.0);
    s := lc + Trunc(((1.0-f) * flabuf[p1, 0]) + (f * flabuf[p2, 0]));
    flabuf[flapos, 0] := lc;
    lc := Clip(s);
    s := rc + Trunc(((1.0-f) * flabuf[p1, 1]) + (f * flabuf[p2, 1]));
    flabuf[flapos, 1] := rc;
    rc := Clip(s);
    d^ := MakeLong(lc, rc);
    Inc(d);
    Inc(flapos);
    if (flapos = FLABUFLEN) then flapos := 0;
    flas := flas + flasinc;
    if (flas < 0) or (flas > FLABUFLEN) then
      flasinc := -flasinc;
    length := length - 4;
  end;
end;

// This procedure is called at the end of playing stream file
procedure SyncProc(SyncHandle : HSYNC; Channel, data, user : DWORD); stdcall;
begin
   PostMessage(User, WM_GetToEnd, 0, 0);
end;

procedure TBASSPlayer.Error(msg: string);
var
  s: string;
begin
  s := msg + #13#10 + '(error code: ' + IntToStr(BASS_ErrorGetCode) + ')';
  MessageBox(Application.handle, PChar(s), 'Error', MB_ICONERROR or MB_OK);
end;

procedure TBASSPlayer.ClearEffectHandle;
var
   i : integer;
begin
   fladspHandle := 0;
   if FDX8EffectReady then
   begin
      for i := 0 to (NumEQBands-1) do
          EQHandle[i] := 0;
      EchoHandle := 0;
      ReverbHandle := 0;
   end;
end;

function TBASSPlayer.IsPluginReady : boolean;
begin
   result := OutputReady;
end;

function TBASSPlayer.PlayLength : DWORD;
var
   SongLength : int64;
   FloatLen : FLOAT;
begin
   result := 0;
   case ChannelType of
      Channel_NotOpened : exit;
      Channel_WMA, Channel_Stream : SongLength := BASS_StreamGetLength(Channel);
      Channel_Music  : SongLength := BASS_MusicGetLength(Channel, true);
      Channel_Plugin : begin
                         result := FStreamInfo.Duration;
                         exit;
                       end;
   end;

   if SongLength < 0 then
      SongLength := 0;

   FloatLen := BASS_ChannelBytes2Seconds(Channel, SongLength);
   result := round(1000 * FloatLen);   // sec -> milli sec
end;


function TBASSPlayer.CurrentPosition : DWORD;
var
   SongPos : int64;
   FloatPos : FLOAT;
begin
   result := 0;
   case ChannelType of
      Channel_NotOpened : exit;
      Channel_WMA, Channel_Stream : SongPos := BASS_ChannelGetPosition(Channel);
      Channel_Music  : exit;   // not implemented yet
      Channel_Plugin : begin
                         if FPlayerMode = plmReady then
                            result := 0
                         else if UsesOutputPlugin then
                            result := GetOutputTime(true)  // Output plug-in's time
                         else
                            result := GetOutputTime(false); // Input plug-in's time
                         exit;
                       end;
   end;

   if SongPos < 0 then
      SongPos := 0;

   FloatPos := BASS_ChannelBytes2Seconds(Channel, SongPos);
   result := round(1000 * FloatPos);     // sec -> milli sec
end;


procedure TBASSPlayer.SetPosition(Value : DWORD);
var
   SongPos : int64;
   wCycle : integer;
begin
   if ChannelType = Channel_NotOpened then
      exit;

   if Value > FStreamInfo.Duration then
      exit;

   if NowStarting then
      exit;

   SongPos := BASS_ChannelSeconds2Bytes(Channel, Value / 1000);
   case ChannelType of
      Channel_NotOpened : exit;
      Channel_Music   : exit;   // not implemented yet
      Channel_WMA, Channel_Stream  : BASS_ChannelSetPosition(Channel, SongPos);
      Channel_Plugin  : begin
                           if (FPlayerMode = plmPlaying) and UsesOutputPlugin then
                              NowStarting := true;
                      //     PostMessage(MessageHandle2, WM_PosChange, 0, 0);
                           SetPosChanged;
                           APlugin.SetOutputTime(Value);
                           wCycle := 0;
                           while FPlayerMode = plmPlaying do
                           begin
                              if NowStarting = false then
                               begin
                                  TimerFFT.Enabled := true;
                                  break;
                               end;
                               WinProcessMessages;
                               sleep(40);
                               inc(wCycle);
                               if wCycle > 50 then   // 2 sec time out
                                  break;
                           end;
                           NowStarting := false;
                        end;
   end;
end;


function TBASSPlayer.GetVolume : integer;
begin
   if ChannelType = Channel_Plugin then
      result := FPluginVolume
   else if ChannelType = Channel_Music then
      result := FMusicVolume
   else
      result := FStreamVolume;
end;

procedure TBASSPlayer.SetVersionStr(Value : string);
begin
   // nothing to do : only to show FVersionStr at form design
end;

procedure TBASSPlayer.SetDLLVersionStr(Value : string);
begin
   // nothing to do : only to show FDLLVersionStr at form design
end;

procedure TBASSPlayer.SetDX8EffectReady(Value : boolean);
begin
   // nothing to do : only to show FDX8EffectReady at form design
end;

procedure TBASSPlayer.SetVolume(Value : integer);
var
   tmpValue : integer;
begin
   if Value > 100 then
      tmpValue := 100
   else if Value < -1 then
      tmpValue := 0
   else
      tmpValue := Value;

 // Adjust the volume according ChannelType
   if ChannelType = Channel_Plugin then
   begin
      APlugin.SetVolume(round(tmpValue * 255.0 / 100.0));
      FPluginVolume := tmpValue;
   end else if ChannelType = Channel_Music then
   begin
      BASS_SetGlobalVolumes(tmpValue, -1, -1);
      FMusicVolume := tmpValue;
   end else begin
      BASS_SetGlobalVolumes(-1, -1, tmpValue);
      FStreamVolume := tmpValue;
   end;
end;

procedure TBASSPlayer.SetEQGains(Value : TEQGains);
var
   i : integer;
begin
   if not FDX8EffectReady then
      exit;

   for i := 0 to (NumEQBands - 1) do
   begin
      if Value[i] > 15.0 then
         FEQGains[i] := 15.0
      else if Value[i] < -15.0 then
         FEQGains[i] := -15.0
      else
         FEQGains[i] := Value[i];

      if not (Equalizer in FSoundEffects) then
         Continue;

      if EQHandle[i] = 0 then
         Continue;

      BASS_FXGetParameters(EQHandle[i], @EQParam);

   // Avoid redundant operation
      if abs(EQParam.fGain - FEQGains[i]) < 0.001 then   // consider round-off error
         Continue;

      EQParam.fBandwidth := FEQBands.BandPara[i].Bandwidth;
      EQParam.fCenter := FEQBands.BandPara[i].CenterFreq;
      EQParam.fGain:= FEQGains[i];
      BASS_FXSetParameters(EQHandle[i], @EQParam);
   end;
end;

procedure TBASSPlayer.SetEchoLevel(Value : word);
begin
   if Value > 32 then
      exit;

   FEchoLevel := Value;
   if Echo in FSoundEffects then
   begin
       if EchoHandle <> 0 then
       begin
         if BASS_FXGetParameters(EchoHandle, @EchoParam) then
         begin
            EchoParam.fWetDryMix := FEchoLevel * 1.2{30.0};
            EchoParam.fFeedBack := 30.0;
            BASS_FXSetParameters(EchoHandle, @EchoParam)
         end;
       end;
   end;
end;

procedure TBASSPlayer.SetReverbLevel(Value : word);
begin
   if Value > 32 then
      exit;

   FReverbLevel := Value;
   if Reverb in FSoundEffects then
   begin
      if ReverbHandle <> 0 then
       begin
         if BASS_FXGetParameters(ReverbHandle, @ReverbParam) then
         begin
            ReverbParam.fInGain := 0.0;
            ReverbParam.fReverbMix := FReverbLevel * 0.5 - 16.0;
            ReverbParam.fReverbTime := 1000.0;
            ReverbParam.fHighFreqRTRatio := 0.1;
            BASS_FXSetParameters(ReverbHandle, @ReverbParam)
         end;
       end;
   end;
end;

procedure TBASSPlayer.SetEQBands(Value : TEQBands);
var
   i : integer;
begin
   FEQBands.Bands := Value.Bands;
   if FEQBands.Bands > NumEQBands then
      FEQBands.Bands := NumEQBands;

   for i := 0 to (FEQBands.Bands-1) do
   begin
      FEQBands.BandPara[i].CenterFreq := Value.BandPara[i].CenterFreq;
      FEQBands.BandPara[i].Bandwidth := Value.BandPara[i].Bandwidth;
   end;

   if Equalizer in FSoundEffects then
     for i := 0 to (FEQBands.Bands-1) do
     begin
        if EQHandle[i] = 0 then
           EQHandle[i] := BASS_ChannelSetFX(Channel, BASS_FX_PARAMEQ);
        EQParam.fGain := FEQGains[i];
        EQParam.fBandwidth := FEQBands.BandPara[i].Bandwidth;
        EQParam.fCenter := FEQBands.BandPara[i].CenterFreq;
        BASS_FXSetParameters(EQHandle[i], @EQParam);
     end;

   for i := FEQBands.Bands to (NumEQBands-1) do
   begin
      if EQHandle[i] <> 0 then
         if BASS_ChannelRemoveFX(Channel, EQHandle[i]) then
            EQHandle[i] := 0;
   end;
end;

procedure TBASSPlayer.SetSoundEffect(Value : TSoundEffects);
var
//   tmpPaused : boolean;
   i : integer;
begin
   FSoundEffects := Value;

   if ChannelType = Channel_NotOpened then
      exit;

   if Flanger in Value then
   begin
      if fladspHandle = 0 then
      begin
         FillChar(flabuf, SizeOf(flabuf), 0);
         flapos := 0;
         flas := FLABUFLEN / 2;
         flasinc := 0.002;
         fladspHandle := BASS_ChannelSetDSP(Channel, Flange, 0);
      end;
   end
   else
      if fladspHandle <> 0 then
         if BASS_ChannelRemoveDSP(Channel, fladspHandle) then
            fladspHandle := 0;

   if not FDX8EffectReady then
      exit;

// DX8 effects can be applied at any time with BASS ver 1.7
{  if BASS_ChannelIsActive(Channel) = BASS_ACTIVE_PLAYING then
   begin
      PausePlay;
      tmpPaused := true;
   end else
      tmpPaused := false; }

   if Equalizer in Value then
   for i := 0 to (FEQBands.Bands-1) do
   begin
      if EQHandle[i] = 0 then
         EQHandle[i] := BASS_ChannelSetFX(Channel, BASS_FX_PARAMEQ);
      EQParam.fGain := FEQGains[i];
      EQParam.fBandwidth := FEQBands.BandPara[i].Bandwidth;
      EQParam.fCenter := FEQBands.BandPara[i].CenterFreq;
      BASS_FXSetParameters(EQHandle[i], @EQParam);
   end else
   for i := 0 to (NumEQBands-1) do
   begin
      if EQHandle[i] <> 0 then
         if BASS_ChannelRemoveFX(Channel, EQHandle[i]) then
            EQHandle[i] := 0;
   end;

   if Echo in Value then
   begin
       if EchoHandle = 0 then
          EchoHandle := BASS_ChannelSetFX(Channel, BASS_FX_ECHO);
       if EchoHandle <> 0 then
       begin
         if BASS_FXGetParameters(EchoHandle, @EchoParam) then
         begin
            EchoParam.fWetDryMix := FEchoLevel * 1.2{30.0};
            EchoParam.fFeedBack := 30.0;
            BASS_FXSetParameters(EchoHandle, @EchoParam)
         end;
       end;
   end else
      if EchoHandle <> 0 then
         if BASS_ChannelRemoveFX(Channel, EchoHandle) then
            EchoHandle := 0;

   if Reverb in Value then
   begin
       if ReverbHandle = 0 then
          ReverbHandle := BASS_ChannelSetFX(Channel, BASS_FX_REVERB);
       if ReverbHandle <> 0 then
       begin
         if BASS_FXGetParameters(ReverbHandle, @ReverbParam) then
         begin
            ReverbParam.fInGain := 0.0;
            ReverbParam.fReverbMix := FReverbLevel * 0.5 - 16.0;
            ReverbParam.fReverbTime := 1000.0;
            ReverbParam.fHighFreqRTRatio := 0.1;
            BASS_FXSetParameters(ReverbHandle, @ReverbParam)
         end;
       end;
   end else
      if ReverbHandle <> 0 then
         if BASS_ChannelRemoveFX(Channel, ReverbHandle) then
            ReverbHandle := 0;

 {  if tmpPaused then
      ResumePlay;  }
end;


function  TBASSPlayer.RunVisoutProgram(VismodName : string; VismodNo : word) : boolean; // * New at Ver 1.3
var
   TimerStat : boolean;
 //  StartupInfo: TStartupInfo;
 //  ProcessInfo: TProcessInformation;
 //  BeginPath: PChar;
begin
 // Winamp visualization plug-ins are designed to be run through their own user
 // interface thread.
 // I have made a program named "VisOut.exe" to run Winamp visualization plug-ins
 // in a seperate process.
 // The visualization data are fed to "VisOut.exe" via shared memory.
 // note : Some Winamp visualization plug-ins won't be run, which require that
 //        Winamp.exe be the creator of them. I do not know the method to
 //        overcome this obstacle.

   result := false;
   if not FileExists(VismodName) then
      exit;
   if FindWindow('TApplication', 'VisOut') <> 0 then  // if VisOut is running, quit it first.
      fisSharedMemory1.InformDataReady(20{QuitProg}, 0);

   TimerStat := TimerFFT.Enabled;
   TimerFFT.Enabled := false;

   AppExec1.ExeName := GetProgDir + 'VisOut.exe "' + VismodName + '" ' + intToStr(VismodNo);
   AppExec1.changeDir := false;
   try
     AppExec1.Execute;
     GoVisOut := true;
     result := true;
   except
     on E:EAppExec do ShowMessage(E.Message);
   end;

  // Following statements(equivalent as AppExec1.Exec) causes error if executed
  // while TBassPlayer is playing.  What is the reason ?
  { FillChar (StartupInfo, Sizeof(StartupInfo), 0);
   StartupInfo.cb := Sizeof(StartupInfo);
   StartupInfo.dwFlags := STARTF_USESHOWWINDOW;
   StartupInfo.wShowWindow := SW_NORMAL;  // SW_HIDE;
   BeginPath := pchar(GetProgDir);
   Result := CreateProcess(nil, pchar(GetProgDir + 'VisOut.exe ' + VismodName + ' ' + intToStr(VismodNo)),
                           nil, nil, False, CREATE_NEW_CONSOLE or NORMAL_PRIORITY_CLASS,
                           nil, BeginPath, StartupInfo, ProcessInfo);

   if Result then
   begin
     CloseHandle(ProcessInfo.hThread);
   //  VisProcHandle := ProcessInfo.hProcess;
     CloseHandle(ProcessInfo.hProcess);
     GoVisOut := true;
   end else
      MessageBox(Application.Handle, 'Failed to run VisOut.exe', 'GetLastError',
                                                   MB_OK or MB_ICONINFORMATION); }

   TimerFFT.Enabled := TimerStat;
end;

procedure TBASSPlayer.QuitVisoutProgram;       // * New at Ver 1.3
begin
   GoVisOut := false;

 // Inform "VisOut.exe" that it's time to end
   if FindWindow('TApplication', 'VisOut') <> 0 then
      fisSharedMemory1.InformDataReady(20{QuitProg}, 0);
end;

procedure TBASSPlayer.SetPluginFirst(Value : boolean);
begin
   FPluginFirst := Value;
end;

// This procedure is used to inhibit or to allow unloading of Winamp
// plug-ins.
//   SetUnloadBut(false) -> inhibit,  SetUnloadBut(true) -> allow
procedure TBASSPlayer.SetUnloadBut(Value : boolean);
begin
   FUnloadButEnabled := value;
   PluginConfigForm.btnUnload.Enabled := value;
end;


constructor TBASSPlayer.Create(AOwner: TComponent);
var
   i : integer;
   MusicVol, SampleVol, StreamVol : DWORD;
   BASSVersion : DWORD;
begin
   inherited Create(AOwner);

   FVersionStr := Ver;
   FBASSReady := false;
   FBASSWMAReady := false;
   ChannelType := Channel_NotOpened;
   FSoundEffects := [];
   FPlayerMode := plmStandby;
   FStreamName :='';
   FStreamInfo.FileName := '';
   FStreamInfo.SampleRate := 0;
   FSupportedBy := None;
   FPluginFirst := false;

   if not (csDesigning in ComponentState) then
   begin
     AppHWND := Application.Handle;
     MessageHandle := AllocateHWnd(ProcMessage);
     SetMainHandle(MessageHandle);  // set hMainWindow in PluginCtrl unit
     TimerFFT := TTimer.Create(nil);
     TimerFFT.Interval := 33;
     TimerFFT.Enabled := false;
     TimerFFT.OnTimer := TimerFFTTimer;
   end;

   BASSDLLLoaded := Load_BASSDLL(GetProgDir, 'bass.dll');
   if not BASSDLLLoaded then
   begin
      if (csDesigning in ComponentState) then
         FDLLVersionStr := 'N.A.(Copy bass.dll in <directory Delphi installed>\Bin to get version)'
      else
         FDLLVersionStr := '';
      exit;
   end;

   BASSVersion := BASS_GetVersion;
   if (BASSVersion < MAKELONG(1,8)) then
   begin
     if not (csDesigning in ComponentState) then
        Error('BASS version is not 1.8 or higher');
     exit;
   end;

 // setup output - default device, 44100hz, stereo, 16 bits
   if not BASS_Init(-1, 44100, 0, AppHWND) then
   begin
      if not (csDesigning in ComponentState) then
         Error('Can''t initialize device');
      exit;
   end;

   if BASS_Start then
      FBASSReady := true
   else
      exit;

   BassInfoParam.size := SizeOf(BassInfoParam);
   BASS_GetInfo(BassInfoParam);

 // BASS_ChannelSetFX requires DirectX version 8 or higher
   if BassInfoParam.dsver >= 8 then
   begin
      FDX8EffectReady := true;
   end else begin
      if not (csDesigning in ComponentState) then
         MessageBox(AppHWND, NoDX8Msg, 'Warning', MB_ICONWARNING or MB_OK);
      FDX8EffectReady := false;
   end;

   FEQBands.Bands := NumEQBands;
   for i := 0 to (NumEQBands-1) do
   begin
      FEQGains[i] := 0;
      FEQBands.BandPara[i].CenterFreq := EQFreq[i];
      FEQBands.BandPara[i].Bandwidth := EQBandWidth[i];
   end;
   FEchoLevel := 16;
   FReverbLevel := 16;

   BASS_GetGlobalVolumes(MusicVol, SampleVol, StreamVol);
   FMusicVolume := MusicVol;
   FStreamVolume := StreamVol;
   FPluginVolume := 255;
   FDLLVersionStr := GetDLLVersionStr;

   if not (csDesigning in ComponentState) then
   begin
      if FBASSReady then
         if Load_BASSWMADLL(GetProgDir, 'basswma.dll') then
            FBASSWMAReady := true;

      InitPluginCtrl;
      MessageHandle2 := GetMsgHandler;  // = message handler of PluginCtrl unit
      MPEG := TMPEGaudio.Create;
      Vorbis := TOggVorbis.Create;
      WMA := TWMAfile.Create;
      WAV := TWAVFile.Create;

  // Is it right to create a form component in this way ?
  // Anyway I do not know the better way and this method works.
      PluginConfigForm := TPluginConfigForm.Create(Self);
      FUnloadButEnabled := PluginConfigForm.btnUnload.Enabled;

      AppExec1 := TAppExec.Create(self);
      fisSharedMemory1 := TfisSharedMemory.Create(self);
      fisSharedMemory1.ShareName := 'fisSharedMemory1';
      fisSharedMemory1.Size := 2312;
      fisSharedMemory1.TargetApplication := 'VisOut';
      if fisSharedMemory1.CreateMemory then
         ShareMemPointer := fisSharedMemory1.MapMemory;
   end;
end;


destructor  TBASSPlayer.Destroy;
begin
   if ChannelType = Channel_Plugin  then
   begin
      if Channel <> 0 then
         APlugin.Stop
   end else if (FPlayerMode = plmPlaying) then
      Stop;

   if BASSDLLLoaded then
   begin
      BASS_Free;
      Unload_BASSDLL;
   end;
   if FBASSWMAReady then
      Unload_BASSWMADLL;

   if not (csDesigning in ComponentState) then
   begin
     TimerFFT.Free;
     MPEG.Free;
     Vorbis.Free;
     WMA.Free;
     WAV.Free;
     QuitPluginCtrl;
     if MessageHandle <> 0 then
        DeallocateHWnd(MessageHandle);
     PluginConfigForm.Free;

     AppExec1.Free;
     if ShareMemPointer <> nil then
      fisSharedMemory1.UnMapMemory(ShareMemPointer);
     fisSharedMemory1.CloseMemory;
     fisSharedMemory1.Free;
   end;

   inherited Destroy;
end;

// Get the version information as string
function TBASSPlayer.GetDLLVersionStr : string;
var
   VersionNum : DWORD;
begin
   if not FBASSReady then
   begin
      result := '';
      exit;
   end;

   VersionNum := BASS_GetVersion;
   result := intToStr(LoWord(VersionNum)) + '.' + intToStr(HiWord(VersionNum));
end;

function TBASSPlayer.GetNativeFileExts : string;
begin
   if FBASSReady and FBASSWMAReady then
      result := '*.wav;*.mp3;*.ogg;*.wma;'
   else if FBASSReady then
      result := '*.wav;*.mp3;*.ogg;'
   else
      result := '';
end;

function TBASSPlayer.GetPluginFileExts : string;
begin
   result := GetPlayableFiles;
end;

function TBASSPlayer.GetStreamInfo2(StreamName : string;
                       var StreamInfo : TStreamInfo;
                       var SupportedBy : TSupportedBy;
                       var PluginNum : integer) : boolean;
var
   i : integer;
   s, ExtCode : string;
   APlugin : TPlugin;
   APluginInfo : TPluginInfo;
   _file : array[0..255] of char;
   _title : array[0..255] of char;
   _length_in_ms : integer;
begin
   result := false;

   with StreamInfo do
   begin
      FileName := '';
      SampleRate := 0;
      BitRate := 0;
      Duration := 0;
      Title := '';
      Artist := '';
      Album := '';
      Year := '';
      Genre := '';
      GenreID := 0;
      Track := 0;
      Comment := '';
      Channels := 0;
   end;

//   StreamInfo.FileName := ExtractFileName(StreamName);
   StreamInfo.FileName := StreamName;  // * Changed at Ver 1.3
   SupportedBy := None;
   PluginNum := -1;
   ExtCode := UpperCase(ExtractFileExt(StreamName));

 // Check if native file types
   if ExtCode = '.WAV' then
   begin
      WAV.ReadFromFile(StreamName);
      if WAV.Valid then
      begin
         with StreamInfo do
         begin
            FileSize := WAV.FileSize;
            SampleRate := WAV.SampleRate;
            Duration := round(WAV.Duration * 1000);
            Channels := WAV.ChannelModeID;
         end;
         if FBASSReady then
            SupportedBy := BASSNative;
         result := true;
      end;
   end else if ExtCode = '.MP3' then
   begin
      MPEG.ReadFromFile(StreamName);
      if MPEG.Valid then
      begin
         with StreamInfo do
         begin
            FileSize := MPEG.FileLength;
            SampleRate := MPEG.SampleRate;
            BitRate := MPEG.BitRate;
            Duration := round(MPEG.Duration * 1000);

            if MPEG.ID3v1.Exists then
            begin
               Title := MPEG.ID3v1.Title;
               Artist := MPEG.ID3v1.Artist;
               Album := MPEG.ID3v1.Album;
               Year := MPEG.ID3v1.Year;
               Genre := MPEG.ID3v1.Genre;
               GenreID := MPEG.ID3v1.GenreID;
               Track := MPEG.ID3v1.Track;
               Comment := MPEG.ID3v1.Comment;
            end else if MPEG.ID3v2.Exists then
            begin
               Title := MPEG.ID3v2.Title;
               Artist := MPEG.ID3v2.Artist;
               Album := MPEG.ID3v2.Album;
               Year := MPEG.ID3v2.Year;
               Genre := MPEG.ID3v2.Genre;
               Track := MPEG.ID3v2.Track;
               Comment := MPEG.ID3v2.Comment;
            end;

            if MPEG.ChannelMode = 'Mono' {MPEG_CM_MONO} then
               Channels := 1
            else
               Channels := 2;
         end;
         if FBASSReady then
            SupportedBy := BASSNative;
         result := true;
      end;
   end else if ExtCode = '.OGG' then
   begin
      Vorbis.ReadFromFile(StreamName);
      if Vorbis.Valid then
      begin
         with StreamInfo do
         begin
            FileSize := Vorbis.FileSize;
            SampleRate := Vorbis.SampleRate;
            BitRate := Vorbis.BitRate;
            Duration := round(Vorbis.Duration * 1000);
            Title := Vorbis.Title;
            Artist := Vorbis.Artist;
            Album := Vorbis.Album;
            Year := Vorbis.Date;
            Genre := Vorbis.Genre;
            Channels := Vorbis.ChannelModeID;
            Track := Vorbis.Track;
            Comment := Vorbis.Comment;
         end;
         if FBASSReady then
            SupportedBy := BASSNative;
         result := true;
      end;
   end else if ExtCode = '.WMA' then
   begin
      WMA.ReadFromFile(StreamName);
      if WMA.Valid then
      begin
         with StreamInfo do
         begin
            FileSize := WMA.FileSize;
            SampleRate := WMA.SampleRate;
            BitRate := WMA.BitRate;
            Duration := round(WMA.Duration * 1000);
            Title := WMA.Title;
            Artist := WMA.Artist;
            Album := WMA.Album;
            Year := WMA.Year;
            Genre := WMA.Genre;
            Channels := WMA.ChannelModeID;
            Track := WMA.Track;
            Comment := WMA.Comment;
         end;
         if FBASSWMAReady then
            SupportedBy := BASSNative;
         result := true;
      end;
   end;

 // Check if it is the file type that can be played by Winamp input plug-in
   for i := 0 to MaxPluginNum - 1 do
   begin
      APlugin := GetPlugin(i);
      if APlugin <> nil then
      begin
         APluginInfo := GetPluginInfo(i);
         s := uppercase(APluginInfo.FileExtensions);
         if pos(copy(ExtCode, 2, 3), s) <> 0 then
         begin
            PluginNum := i;
            StrPCopy(_file, StreamName);
            APlugin.GetFileInfo(_file, _title, _length_in_ms);
            if (_length_in_ms > 0) then
            begin
               with StreamInfo do
               begin
                  if SupportedBy = BASSNative then
                     SupportedBy := Both
                  else begin
                     if FBASSReady then
                        SupportedBy := WinampPlugin;
                     Title := string( _title);
                     Duration := _length_in_ms;
                     result := true;
                  end;
                  break;
               end;
            end;
         end;
      end;
   end;
end;

function TBASSPlayer.GetStreamInfo(StreamName : string;
                       var StreamInfo : TStreamInfo;
                       var SupportedBy : TSupportedBy) : boolean;
var
   dumNum : integer;
begin
   result := GetStreamInfo2(StreamName, StreamInfo, SupportedBy, dumNum);
end;


procedure TBASSPlayer.ProcMessage(var Msg: TMessage);
var
   PChannelInfo : ^TChannelInfo;
begin
   case Msg.Msg of
      WM_NewFFTData    : if Assigned(FOnNewFFTData) then
                            FOnNewFFTData(Self, BandOut);
   // We do not know the Channels & SampleRate of opened stream file before Winamp
   // input plug-in informs to output plug-in, if the file type is the one which is only
   // supported by Winamp input plug-in.
   // WM_StartPlay message is used to receive the Channels & SampleRate of opened
   // stream file at starting play from output plug-in emulator.
   // We still cannot know the Channels & SampleRate of opened stream file if its related
   // Winamp input plug-in is the type which directly sound out (= not thru output plug-in emulator).
      WM_StartPlay     : begin                       // Message from output plug-in emulator
                            NowStarting := false;
                            UsesOutputPlugin := true;
                            PChannelInfo := pointer(Msg.lParam);
                            FStreamInfo.Channels := PChannelInfo^.Channels;
                            FStreamInfo.SampleRate := PChannelInfo^.SampleRate;
                            if Assigned(OnPluginStartPlay) then
                               FOnPluginStartPlay(Self, FStreamInfo.Channels,
                                                        FStreamInfo.SampleRate);
                            if Assigned(FOnNewFFTData) then
                               TimerFFT.Enabled := true;
                            SetSoundEffect(FSoundEffects);
                         end;
      WM_BASS_StreamCreate : Channel := Msg.WParam;  // Message from output plug-in emulator
     { WM_BASS_StreamFree   : begin                   // Message from output plug-in emulator
                                Channel := 0;
                                ClearEffectHandle;
                             end; }
      WM_WA_MPEG_EOF   : begin                       // Message from Winamp input plug-in emulator
                         //   SetPlayEnded;
                            NowStarting := false;
                            if UsesOutputPlugin then
                               while BASS_ChannelIsActive(Channel) = BASS_ACTIVE_PLAYING do
                               begin
                                  WinProcessMessages;
                                  sleep(50);
                               end;

                            Stop;
                            if Assigned(FOnPlayEnd) then
                               FOnPlayEnd(Self);
                         end;
      WM_GetToEnd      : begin
                            FPlayerMode := plmStopped;
                            if Assigned(FOnPlayEnd) then
                               FOnPlayEnd(Self);
                         end;
      WM_PluginFirst_Changed : if Msg.lParam = 0 then  // Message from PluginConfigForm
                                  FPluginFirst := false
                               else
                                  FPluginFirst := true;
      WM_QueryEndSession : Msg.Result := 1;    // Allow system termination
   end;
end;


function TBASSPlayer.Open(StreamName : string) : boolean;
var
   StreamInfo : TStreamInfo;
   SupportedBy : TSupportedBy;
   PluginNum : integer;
   chattr: Integer;
begin
   result := false;

   if not FBASSReady then
   begin
      Application.MessageBox('Player is not ready !', 'Confirm', MB_OK);
      exit;
   end;

   if not GetStreamInfo2(StreamName, StreamInfo, SupportedBy, PluginNum) then
   begin
      Error('Invalid stream file(' + StreamName + ') !');
      exit;
   end;

   // release channel
   case ChannelType of
      Channel_Music : BASS_MusicFree(Channel);
      Channel_WMA, Channel_Stream : BASS_StreamFree(Channel);
      Channel_Plugin : if Channel <> 0 then
                          APlugin.Stop;
   end;

   if (SupportedBy = WinampPlugin) or
      ((SupportedBy = Both) and FPluginFirst) then
   begin
      ChannelType := Channel_Plugin;
      FPluginNum := PluginNum;
      APlugin := GetPlugin(PluginNum);
  //    PostMessage(MessageHandle2, WM_StreamOpen, MessageHandle, FPluginNum);
      SelectInputPlugin(FPluginNum);
   end else
   begin
      Channel := BASS_StreamCreateFile(FALSE, PChar(StreamName), 0, 0, 0{BASS_SAMPLE_FX});
      if (Channel <> 0) then
         ChannelType := Channel_Stream
      else
      begin
         if FBASSWMAReady then
            Channel := BASS_WMA_StreamCreateFile(FALSE, PChar(StreamName), 0, 0, 0{BASS_SAMPLE_FX}) // it's a WMA?
         else
            Channel := 0;
         if (Channel <> 0) then
            ChannelType := Channel_WMA
         else begin
            Channel := BASS_MusicLoad(FALSE, PChar(StreamName), 0, 0, BASS_MUSIC_CALCLEN);  // * changed at Ver 1.3 (added flag BASS_MUSIC_CALCLEN)
            if (Channel <> 0) then
               ChannelType := Channel_Music
            else
               ChannelType := Channel_NotOpened;
         end;
      end;

      if (Channel = 0) then  // not a playable file
      begin
         Error('Can''t play the file');
         Exit;
      end;

      chattr := BASS_ChannelGetFlags(Channel) and ({BASS_SAMPLE_MONO or} BASS_SAMPLE_8BITS);
      if chattr > 0 then
      begin
      // does not support 16-bit sources.
         BASS_MusicFree(Channel);
         BASS_StreamFree(Channel);
         Error('Sound card does not support 16-bit sources.');
         Exit;
      end;
   end;

  // Reset previous handle values
   ClearEffectHandle;

  // Apply equalyzation and sound effects
   if ChannelType <> Channel_Plugin then
      SetSoundEffect(FSoundEffects);

   FStreamName := StreamName;
   FStreamInfo := StreamInfo;
   FSupportedBy := SupportedBy;
   FPlayerMode := plmReady;

   result := true;
end;


procedure TBASSPlayer.Play;
var
   StartOK : boolean;
   wCycle : integer;
begin
   if NowStarting then   // avoid duplicate Play operation
      exit
   else if ChannelType = Channel_NotOpened then
      exit
   else if FPlayerMode = plmPlaying then
      exit
   else if FPlayerMode = plmPaused then
   begin
      ResumePlay;
      exit;
   end else if FPlayerMode = plmStopped then
   begin
      if ChannelType <> Channel_Plugin then
         SetPosition(0);   // Set position to starting point
      Restart;
      exit;
   end;

   StartOK := false;

 // play an opened channel
   case ChannelType of
      Channel_Music : StartOK := BASS_MusicPlay(Channel);
      Channel_WMA, Channel_Stream : StartOK := BASS_StreamPlay(Channel, false, 0);
      Channel_Plugin :  begin
                            NowStarting := true;
                            UsesOutputPlugin := false;
                            Channel := 1;  // for Winamp input plug-in which does not need output plug-in
                            APlugin.Play(pchar(FStreamName));
                            SetVolume(FPluginVolume);
                            wCycle := 0;
                            while wCycle < 60 do
                            begin
                               if GetOutputTime(false) > 200 then
                               begin
                                  StartOK := true;
                                  break;
                               end;
                               WinProcessMessages;
                               inc(wCycle);
                               sleep(50);
                            end;
                            NowStarting := false;
                        end;
   end;

   if StartOK then
   begin
      if ChannelType = Channel_WMA then
          SyncHandle := BASS_WMA_ChannelSetSync(Channel, BASS_SYNC_END, 0, @SyncProc, MessageHandle)
       else if ChannelType <> Channel_Plugin then
          SyncHandle := BASS_ChannelSetSync(Channel, BASS_SYNC_END, 0, @SyncProc, MessageHandle);
      FPlayerMode := plmPlaying;
      if (ChannelType <> Channel_Plugin) then
         if Assigned(FOnNewFFTData) then
            TimerFFT.Enabled := true;
   end;
end;

procedure TBASSPlayer.PausePlay;
begin
   if FPlayerMode <> plmPlaying then
      exit;

   if ChannelType = Channel_Plugin then
      APlugin.Pause
   else
      BASS_ChannelPause(Channel);

   FPlayerMode := plmPaused;
end;

procedure TBASSPlayer.ResumePlay;
begin
   if FPlayerMode <> plmPaused then
      exit;

   if ChannelType = Channel_Plugin then
      APlugin.UnPause
   else
      BASS_ChannelResume(Channel);

   FPlayerMode := plmPlaying;
   if Assigned(FOnNewFFTData) then
      if ChannelType <> Channel_Plugin then
         TimerFFT.Enabled := true
      else if UsesOutputPlugin then
         TimerFFT.Enabled := true;
end;

procedure TBASSPlayer.Pause(pAction : boolean);
begin
   if pAction then
      PausePlay
   else
      ResumePlay;
end;

procedure TBASSPlayer.Stop;
begin
   if FPlayerMode <> plmPlaying then
      exit;

   if ChannelType = Channel_Plugin then
   begin
      if Channel <> 0 then
      begin
         APlugin.Stop;
         Channel := 0;
         ClearEffectHandle;
      end
   end else
      BASS_ChannelStop(Channel);

   FPlayerMode := plmStopped;
end;

procedure TBASSPlayer.Restart;
var
   StartOK : boolean;
   wCycle : integer;
begin
   if FPlayerMode <> plmStopped then
      exit;
   if NowStarting then
      exit;
   StartOK := false;

  // play a opened channel
   case ChannelType of
       Channel_Music : StartOK := BASS_MusicPlay(Channel);
       Channel_WMA, Channel_Stream : StartOK := BASS_StreamPlay(Channel, false, 0);
       Channel_Plugin : begin
                           NowStarting := true;
                           UsesOutputPlugin := false;
                           Channel := 1;  // for Winamp input plug-in which does not need output plug-in
                           APlugin.Play(pchar(FStreamName));
                           SetVolume(FPluginVolume);
                           wCycle := 0;
                           while wCycle < 60 do
                           begin
                               if GetOutputTime(false) > 200 then
                               begin
                                  StartOK := true;
                                  break;
                               end;
                               WinProcessMessages;
                               inc(wCycle);
                               sleep(50);
                           end;
                           NowStarting := false;
                        end;
   end;

   if StartOK then
   begin
      FPlayerMode := plmPlaying;
      if (ChannelType <> Channel_Plugin) then
         if Assigned(FOnNewFFTData) then
            TimerFFT.Enabled := true;
   end;
end;


procedure TBASSPlayer.Close;
begin
   case ChannelType of
      Channel_Music : BASS_MusicFree(Channel);
      Channel_Stream, Channel_WMA : BASS_StreamFree(Channel);
      Channel_Plugin : begin
                          if Channel <> 0 then
                          begin
                             APlugin.Stop;
                             Channel := 0;
                             ClearEffectHandle;
                          end;
                          FPluginNum := -1;
                          APlugin := nil;
                          PostMessage(MessageHandle2, WM_StreamClose, 0, 0);
                       end;
   end;

   ChannelType := Channel_NotOpened;
   FPlayerMode := plmStandby;
   FStreamName :='';
   FSupportedBy := None;
   FStreamInfo.FileName := '';
   FStreamInfo.SampleRate := 0;
end;


function TBASSPlayer.GetChannelFFTData(PFFTData : pointer; FFTFlag : DWORD) : boolean;
begin
   result := false;

   if BASS_ChannelIsActive(Channel) <> BASS_ACTIVE_PLAYING then
      exit;

   if (FFTFlag = BASS_DATA_FFT512) or (FFTFlag = BASS_DATA_FFT1024) or
      (FFTFlag = BASS_DATA_FFT2048) or (FFTFlag = BASS_DATA_FFT512S) or
      (FFTFlag = BASS_DATA_FFT1024S) or (FFTFlag = BASS_DATA_FFT2048S) then
      if BASS_ChannelGetData(Channel, PFFTData, FFTFlag) <> $ffffffff then
         result := true;
end;


// If you have any idea on spectrum visualization for better result,
// please let me know.
procedure TBASSPlayer.TimerFFTTimer(Sender: TObject);
const
 // Classify FFTData[x] by the number of required bands for spectrum visualization.
 // FreqCoef defines the last index number of FFTData[x] for each band.
   FreqCoef : array[0..NumFFTBands - 1] of word =
               (  1,  2,    3,   6,       // narrow interval for low freq. range
                 12,  18,  24,  30,  36,  // normal interval for middle freq. range
                 42,  48,  54,  60,  66,
                 72,  78,  84,  90,  96,
                102, 108,
                120, 132, 156, 180 );     // wide interval for high freq. range

  // If the sampling rate is (standard) 44100Hz then the frequency of FFTData[x]
  // is 44100 / 512 * x, so the representative(or maximum) frequency of each
  // band is as followings
  //  BandOut[0] : 86.1Hz     BandOut[1] : 172.3Hz    BandOut[2] : 344.5Hz
  //  BandOut[3] : 516.8Hz    BandOut[4] : 1033.6Hz   BandOut[5] : 1550.4Hz
  //  . . . (516.8Hz interval upto BandOut[20]) . .   BandOut[20] : 9302.2Hz
  //  BandOut[21] : 10336Hz   BandOut[22] : 11370Hz   BandOut[23] : 13437Hz
  //  BandOut[24] : 15503Hz   . . very high frequency range is excluded

   Boost = 0.15; // Factor to intensify high frequency bands
   Scale = 80;   // Factor to calibrate the output range of BandOut[x]
                 //   ( approximately from 0 to 24 ).
var
   FFTData : TFFTData;
   NewBandOut : TBandOut;
   StartIndex : integer;
   i, k : integer;
   tmpIntensity : double;

   SpectrumData: array[0..2048] of float;
   WaveformData: array[0..1151] of DWORD;
   p1 : PDWORD;
   p2 : PBYTE;

   freq, dum1 : dword;
   dum : integer;
begin
   if BASS_ChannelIsActive(Channel) <> BASS_ACTIVE_PLAYING then
   begin
      TimerFFT.Enabled := false;
      exit;
   end;

   if BASS_ChannelGetData(Channel, @FFTData, BASS_DATA_FFT512) = $ffffffff then
   begin
      for i := 0 to (NumFFTBands - 1) do
         BandOut[i] := 0;
      exit;
   end;

   TimerFFT.Enabled := false;  // Avoid duplicate timer interrupt

   for i := 0 to (NumFFTBands - 1) do
   begin
       if i = 0 then
          StartIndex := 1
       else
          StartIndex := FreqCoef[i-1] + 1;

    // tmpIntensity <= the greatest value in classified FFTData[x]'s
       tmpIntensity := 0;
       for k := StartIndex to FreqCoef[i] do
           if FFTData[k] > tmpIntensity then
              tmpIntensity := FFTData[k];

       NewBandOut[i] := round(tmpIntensity * (1 + i * Boost) * Scale);

    // Decrease the value of BandOut[i] smoothly for better looking
       if NewBandOut[i] > BandOut[i] then
          BandOut[i] := NewBandOut[i]
       else
          if BandOut[i] >= 2 then
             dec(BandOut[i], 2)
          else
             BandOut[i] := 0;

    // for the case that original NewBandOut[i] is smaller than BandOut[i] by 1
       if NewBandOut[i] > BandOut[i] then
          BandOut[i] := NewBandOut[i];
   end;

 //  if not GoVisOut then
      PostMessage(MessageHandle, WM_NewFFTData, 0, 0);

   if not GoVisOut then
   begin
      TimerFFT.Enabled := true;
      exit;
   end;

   BASS_ChannelGetAttributes(Channel, freq, dum1, dum);
   p1 := ShareMemPointer;
   p1^ := freq;
   inc(p1, 1);
   p1^ := 2;  // Channels
   inc(p1, 1);
   p2 := pointer(p1);

   FillChar(SpectrumData, SizeOf(SpectrumData), 0);
   FillChar(WaveformData, SizeOf(waveformData), 0);

 // for stereo FFT, use BASS_DATA_FFT2048S instead of BASS_DATA_FFT2048
 //   bass_channelgetdata(Channel, @spectrumdata, BASS_DATA_FFT2048S);
   bass_channelgetdata(Channel, @spectrumdata, BASS_DATA_FFT2048);
   for i := 1 To 576 do
   begin
     p2^ := trunc(SpectrumData[i] * 255);
     inc(p2);
   end;
   for i := 1 To 576 do
   begin
     p2^ := trunc(SpectrumData[i] * 255);
     inc(p2);
   end;

   dum1 := BASS_ChannelGetFlags(Channel);
   if ((dum1 and BASS_SAMPLE_8BITS) = BASS_SAMPLE_8BITS) or
      ((dum1 and BASS_SAMPLE_MONO) = BASS_SAMPLE_MONO) then
   begin
     bass_channelgetdata(Channel, @waveformdata, 1152{=576*2});
     for i := 0 to 575 do
     begin
       p2^ := waveformdata[i] div 256;
       inc(p2);
     end;
     for i := 0 to 575 do
     begin
       p2^ := waveformdata[i] div 256;
       inc(p2);
     end;
   end else
   begin
      bass_channelgetdata(Channel, @waveformdata, 2304{=576*4});
      for i := 0 to 575 do
      begin
        p2^ := waveformdata[i*2] div 256;
        inc(p2);
      end;
      for i := 0 to 575 do
      begin
        p2^ := waveformdata[i*2+1] div 256;
        inc(p2);
      end;
   end;

 // Inform "VisOut.exe" that visualization data is ready
   fisSharedMemory1.InformDataReady(10{DataReady}, 2312{Data Size});

   TimerFFT.Enabled := true;
end;


procedure TBASSPlayer.ShowPluginConfigForm;
var
   SearchRec : TSearchRec;
begin
 // How do we catch the values of parent's elements in the program of child Form ?
 // I cannot implement to get the values of TBASSPlayer's elements in the program
 // of PluginConfigForm.OnShow event procedure.
 // So I have to write the program in the way that TBASSPlayer hands over the
 // values of its elements to PluginConfigForm's.
   if not FBASSReady then
   begin
      Application.MessageBox('Player is not ready !', 'Confirm', MB_OK);
      exit;
   end;

   if FPluginFirst then
      PluginConfigForm.PluginFirst.Checked := true
   else
      PluginConfigForm.PluginFirst.Checked := false;

   PluginConfigForm.PluginList.Clear;
   if FindFirst(GetProgDir + 'Plugins\in_*.dll', 0, SearchRec) = 0 then
   begin
      PluginConfigForm.PluginList.Items.Add(SearchRec.Name);
      while FindNext(SearchRec) = 0 do
         PluginConfigForm.PluginList.Items.Add(SearchRec.Name);
   end;
   FindClose(SearchRec);

   with PluginConfigForm.PluginGrid do
   begin
      Cells[0, 0] := 'Name';
      Cells[1, 0] := 'Decsription';
      Cells[2, 0] := 'Playable File Types';
   end;
   PluginConfigForm.ClearPluginGrid;
   PluginConfigForm.FillPluginGrid;
   SetMessageHandle(MessageHandle);
   PluginConfigForm.ShowModal;
end;

procedure Register;
begin
   RegisterComponents('Samples', [TBASSPlayer]);
end;

end.
