// Unit BASSTest Ver 1.3
//   The program to test TBASSPlayer commponent
//
//    written by Silhwan Hyun  hsh@chollian.net
//
// Note : bass.dll & basswma.dll should be copied in the directory where executable
//        (compiled) program resides in and Winamp 2 input plug-ins in the
//        directory <directory where executable program resides in>\Plugins
//
// Ver 1.3                         5 Apr 2003
//  - Added a button to invoke Winamp visualization plug-in control form, which
//    enables you to select, start and stop visualization plug-ins.
//  - Added two sliders to adjust the depth of echo & reverb effects
//
// Ver 1.2                         5 Feb 2003
//  - Added a button to invoke Winamp input plug-in configuration form to play
//    stream files which BASS sound system cannot decode.
//  - Compiled with Delphi 4
//
// Ver 1.1                         8 Dec 2002
//   - Compiled on Delphi 4 for old version Delphi users
//   - Modified procedure DisplayFFTBand for better result
//
// Ver 1.0                         4 Aug 2002
//   - Initial release
//   - Compiled with Delphi 6

unit BassTest;

interface

uses
  Windows, Messages, SysUtils, {Variants,} Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, BassPlayer, PluginCtrl, Knob, slider, ExtCtrls,
  appexec;

type
  TMainForm = class(TForm)
    btnOpen: TButton;
    OpenDialog1: TOpenDialog;
    btnPlay: TButton;
    btnPause: TButton;
    btnStop: TButton;
    btnClose: TButton;
    VolumeKnob: TKnob;
    PosSlider: TSlider;
    cbFlanger: TCheckBox;
    cbEcho: TCheckBox;
    cbReverb: TCheckBox;
    Slider2: TSlider;
    Slider3: TSlider;
    Slider4: TSlider;
    Slider5: TSlider;
    Slider6: TSlider;
    Slider7: TSlider;
    Slider8: TSlider;
    Slider9: TSlider;
    cbEqualizer: TCheckBox;
    Timer_Stat: TTimer;
    Label_State: TLabel;
    Label_Length: TLabel;
    Slider1: TSlider;
    Slider10: TSlider;
    Label1: TLabel;
    Label2: TLabel;
    Panel1: TPanel;
    GaugePaintBox: TPaintBox;
    Label3: TLabel;
    Label_Title: TLabel;
    InfoMemo: TMemo;
    btnPluginSetup: TButton;
    EchoSlider: TSlider;
    ReverbSlider: TSlider;
    btnShowVisCtrlForm: TButton;

    procedure FormShow(Sender: TObject);
    procedure btnOpenClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure btnPlayClick(Sender: TObject);
    procedure btnPauseClick(Sender: TObject);
    procedure btnStopClick(Sender: TObject);
    procedure btnCloseClick(Sender: TObject);
    procedure cbFlangerClick(Sender: TObject);
    procedure cbEchoClick(Sender: TObject);
    procedure cbReverbClick(Sender: TObject);
    procedure VolumeKnobChange(Sender: TObject);
    procedure EQSliderChange(Sender: TObject);
    procedure cbEqualizerClick(Sender: TObject);
    procedure PosSliderStartTracking(Sender: TObject);
    procedure PosSliderStopTracking(Sender: TObject);
    procedure Timer_StatTimer(Sender: TObject);
    procedure CreateBasicImage;
    procedure ShowBackground;
    procedure PluginStartPlay(Sender: TObject; Channels : WORD; SampleRate : DWORD);
    procedure DisplayFFTBand(Sender: TObject; Bands : TBandOut);
    procedure btnPluginSetupClick(Sender: TObject);
    procedure EchoSliderChange(Sender: TObject);
    procedure ReverbSliderChange(Sender: TObject);
    procedure btnShowVisCtrlFormClick(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var
  MainForm: TMainForm;
  BassPlayer1 : TBassPlayer;

implementation

uses VisPluginCtrl;


{$R *.dfm}

const
// Constants for frequency spectrum visualizzer
   BlockWidth = 3;
   HBlockGap = 1;
   HBlockCount = NumFFTBands;
   VLimit = 24;    // Intensity range : 0 ~ 24
   BackColor = clBlack;

var
   SPS : LongInt;    // sample rate
   NowTracking : boolean;
   BasicBMP : TBitMap;
   GaugeTempBMP : TBitMap;
   DisplayBar: TBitmap;
   GaugeRect : TRect;
   EQGains : TEQGains;
   PeakValue : array[1..NumFFTBands] of single;
   PassedCounter : array[1..NumFFTBands] of integer;

procedure TMainForm.FormShow(Sender: TObject);
var
   i : integer;
begin
   BassPlayer1 := TBassPlayer.Create(Self);
   if BassPlayer1.BASSDLLVer = '' then
   begin
      Application.MessageBox('BASS.DLL is not loaded.', 'Confirm', MB_OK);
      exit;
   end;
   if not BassPlayer1.BASSWMAReady then
      Application.MessageBox('BASSWMA.DLL is not loaded.', 'Confirm', MB_OK);

   BassPlayer1.OnNewFFTData := DisplayFFTBand;
   BassPlayer1.OnPluginStartPlay := PluginStartPlay;

   VolumeKnob.Position := BassPlayer1.Volume;
   CreateBasicImage;
   ShowBackground;
   for i := 1 to NumFFTBands do
      PeakValue[i] := 0;

   EQGains[0] := Slider1.Value - 15.0;
   EQGains[1] := Slider2.Value - 15.0;
   EQGains[2] := Slider3.Value - 15.0;
   EQGains[3] := Slider4.Value - 15.0;
   EQGains[4] := Slider5.Value - 15.0;
   EQGains[5] := Slider6.Value - 15.0;
   EQGains[6] := Slider7.Value - 15.0;
   EQGains[7] := Slider8.Value - 15.0;
   EQGains[8] := Slider9.Value - 15.0;
   EQGains[9] := Slider10.Value - 15.0;

   EchoSlider.Value := BassPlayer1.EchoLevel;
   ReverbSlider.Value := BassPlayer1.ReverbLevel;

   if BassPlayer1.CanUsePlugin then
   begin
      LoadWinampPlugin('in_cdda.dll');
      LoadWinampPlugin('in_asf.dll');
      LoadWinampPlugin('in_midi.dll');
   end;

   if BassPlayer1.PlayerReady then
      Timer_Stat.Enabled := true;
end;

procedure TMainForm.btnOpenClick(Sender: TObject);
var
   s1, s2 : string;
begin
   if not BassPlayer1.PlayerReady then
   begin
      Application.MessageBox('Cannot open file (BassPlayer is not ready)',
                 'Confirm', MB_OK);
      exit;
   end;

   OpenDialog1.FileName := '';
   s1 := BassPlayer1.NativeFileExts;
   s2 := BassPlayer1.PluginFileExts;
   OpenDialog1.Filter := 'All playable files |' + s1 + s2 + '|' +
                         'BASS native files (' + s1 + ')|' + s1 + '|' +
                         'Plug-in supported files (' + s2 + ')|' + s2;

   if OpenDialog1.Execute then
      if BassPlayer1.Open(OpenDialog1.FileName) then
      begin
         Label_Length.Caption := 'Length ' +
            FormatDateTime ('nn:ss', BassPlayer1.PlayLength / (1000 * 24 * 60 * 60));
         Label_Title.Caption := BassPlayer1.StreamInfo.Title;
         InfoMemo.Clear;
         InfoMemo.Lines.Add(ExtractFileName(BassPlayer1.StreamInfo.FileName));
         SPS := BassPlayer1.StreamInfo.SampleRate;
         if SPS <> 0 then
         begin
            s1 := intToStr(SPS) + 'Hz';
            if BassPlayer1.StreamInfo.Channels = 1 then
               S2 := 'Mono'
            else
               s2 := 'Stereo';
            InfoMemo.Lines.Add(s1 + ', ' + s2);
         end;

         if (BassPlayer1.SupportedBy = WinampPlugin) or
             ((BassPlayer1.SupportedBy = Both) and BassPlayer1.PluginFirst) then
            InfoMemo.Lines.Add('Playing by Winamp Plug-in');
      // Following statements are only to scroll down Memo1's contents, i.e.,
      //  to make Memo1's 1st line as the top line.
         InfoMemo.SetFocus;
         InfoMemo.SelStart := 0;
         InfoMemo.SelLength := 0;

         BassPlayer1.Play;
      end;
end;

procedure TMainForm.FormClose(Sender: TObject; var Action: TCloseAction);
begin
   BassPlayer1.QuitVisoutProgram;
   BassPlayer1.Free;
   BasicBMP.Free;
   GaugeTempBMP.Free;
   DisplayBar.Free;
end;

procedure TMainForm.btnPlayClick(Sender: TObject);
begin
   BassPlayer1.Play;
  { if BassPlayer1.Mode = plmPlaying then
      Timer_Stat.Enabled := true;  }
end;

procedure TMainForm.btnPauseClick(Sender: TObject);
begin
   BassPlayer1.Pause(true);
end;

procedure TMainForm.btnStopClick(Sender: TObject);
begin
   BassPlayer1.Stop;
end;

procedure TMainForm.btnCloseClick(Sender: TObject);
begin
   Close;
end;

procedure TMainForm.cbFlangerClick(Sender: TObject);
begin
   if cbFlanger.Checked then
      BassPlayer1.SoundEffects := BassPlayer1.SoundEffects + [Flanger]
   else
      BassPlayer1.SoundEffects := BassPlayer1.SoundEffects - [Flanger];

end;

procedure TMainForm.cbEchoClick(Sender: TObject);
begin
   if cbEcho.Checked then
      BassPlayer1.SoundEffects := BassPlayer1.SoundEffects + [Echo]
   else
      BassPlayer1.SoundEffects := BassPlayer1.SoundEffects - [Echo];
end;

procedure TMainForm.cbReverbClick(Sender: TObject);
begin
   if cbReverb.Checked then
      BassPlayer1.SoundEffects := BassPlayer1.SoundEffects + [Reverb]
   else
      BassPlayer1.SoundEffects := BassPlayer1.SoundEffects - [Reverb];
end;


procedure TMainForm.cbEqualizerClick(Sender: TObject);
begin
   if cbEqualizer.Checked then
      BassPlayer1.SoundEffects := BassPlayer1.SoundEffects + [Equalizer]
   else
      BassPlayer1.SoundEffects := BassPlayer1.SoundEffects - [Equalizer];
end;

procedure TMainForm.VolumeKnobChange(Sender: TObject);
begin
   BassPlayer1.Volume := VolumeKnob.Position;
end;

procedure TMainForm.EQSliderChange(Sender: TObject);
var
   BandNum : integer;
begin
   BandNum := (Sender as TSlider).Tag - 1;
   EQGains[BandNum] := (Sender as TSlider).Value - 15.0;
   BassPlayer1.EQGains := EQGains;
end;


procedure TMainForm.PosSliderStartTracking(Sender: TObject);
begin
   NowTracking := true;
end;

procedure TMainForm.PosSliderStopTracking(Sender: TObject);
var
   SongPos : int64;
begin
   SongPos := Trunc(PosSlider.Value * BassPlayer1.PlayLength / PosSlider.MaxValue);
   BassPlayer1.Position := SongPos;

   NowTracking := false;
end;

procedure TMainForm.Timer_StatTimer(Sender: TObject);
begin
   if BassPlayer1.Mode = plmStandby then
      Label_State.Caption := 'Standby'
   else if BassPlayer1.Mode = plmReady then
      Label_State.Caption := 'Ready to play'
   else if BassPlayer1.Mode = plmPlaying then
      Label_State.Caption := 'Playing '
                 +  FormatDateTime ('nn:ss', BassPlayer1.Position / (1000 * 24 * 60 * 60))
   else if BassPlayer1.Mode = plmStopped then
      Label_State.Caption := 'Stopped'
   else if BassPlayer1.Mode = plmPaused then
      Label_State.Caption := 'Paused';

   if BassPlayer1.Mode <> plmPlaying then
   begin
   //   Timer_Stat.Enabled := false;
      ShowBackground;
   end;

   if BassPlayer1.Mode = plmStandby then
      exit;

   if not NowTracking and (BassPlayer1.Mode <> plmStopped) then
      PosSlider.Value := (BassPlayer1.Position * PosSlider.MaxValue) div BassPlayer1.PlayLength;
end;


// Set background image & bar image for spectrum display
procedure TMainForm.CreateBasicImage;
var
   i : integer;
   R, G, B: Integer;
begin
   BasicBMP := TBitMap.Create;
   BasicBMP.Width := (BlockWidth + HBlockGap) * HBlockCount - HBlockGap + 3;
   BasicBMP.Height := VLimit + 1;
   GaugePaintBox.Width := BasicBMP.Width;
   GaugePaintBox.Height := BasicBMP.Height;
   GaugeTempBMP := TBitMap.Create;
   GaugeTempBMP.Width := BasicBMP.Width;
   GaugeTempBMP.Height := BasicBMP.Height;
   GaugeRect.Left := 0;
   GaugeRect.Top := 0;
   GaugeRect.Right := BasicBMP.Width;
   GaugeRect.Bottom := BasicBMP.Height;

   with BasicBMP.Canvas do
   begin
      Brush.Color := BackColor;
      Brush.Style := bsSolid;
      FillRect(Rect(0, 0, BasicBMP.Width, BasicBMP.Height));
      Pen.Color := clBlue;
      Pen.Width := 1;
      Pen.Style := psDot;
      MoveTo(0, 0);
      LineTo(0, BasicBMP.Height);        // Draw Y-axis line
      MoveTo(0, BasicBMP.Height - 1);
      LineTo(BasicBMP.Width, BasicBMP.Height - 1); // Draw X-axis line
   end;

   DisplayBar := TBitmap.Create;
   DisplayBar.PixelFormat := pf32bit;
   DisplayBar.Width := BlockWidth;
   DisplayBar.Height := VLimit;

   R := 255;
   G := 0;
   B := 0;

   for i := 0 to VLimit - 1 do
   begin
     if i > VLimit / 2 then
        Dec(R, Trunc(256 / VLimit))
     else
        Inc(G, Trunc(768 / VLimit));
     if R < 0 then R := 0;
     if G > 255 then G := 255;
     DisplayBar.Canvas.Brush.Color := TColor(RGB(R, G, B));
     DisplayBar.Canvas.FillRect(Rect(0, i, BlockWidth, i + 1));
   end;

end;

procedure TMainForm.PluginStartPlay(Sender: TObject; Channels : WORD; SampleRate : DWORD);
var
   s1, s2 : string;
begin
   if SPS = 0 then     // = have not gotten stream information yet
      begin
         SPS := SampleRate;
         if SPS <> 0 then
         begin
            s1 := intToStr(SampleRate) + 'Hz';
            if Channels = 1 then
               S2 := 'Mono'
            else
               s2 := 'Stereo';
            InfoMemo.Lines.Add(s1 + ', ' + s2);
         end;
      end;
end;

// Display spectrum image
procedure TMainForm.DisplayFFTBand(Sender: TObject; Bands : TBandOut);
var
   tmpRect, BarRect : TRect;
   j : integer;
begin
 // To prevent flickering, use temporary image buffer and process as follows
 // 1) Draw output image on temporary image buffer.
 // 2) Copy temporary image buffer's image to display canvas

 // Copy BasicBMP's image to GaugeTempBMP(used as temporary image buffer)
   BitBlt(GaugeTempBMP.Canvas.Handle, // handle to destination device context
          GaugeRect.Left,	// x-coordinate of destination rectangle's upper-left corner
          GaugeRect.Top,	// y-coordinate of destination rectangle's upper-left corner
          GaugeTempBMP.Width,	// width of destination rectangle
          GaugeTempBMP.Height,	// height of destination rectangle
          BasicBMP.Canvas.Handle, // handle to source device context
          GaugeRect.Left,	// x-coordinate of source rectangle's upper-left corner
          GaugeRect.Top,	// y-coordinate of source rectangle's upper-left corner
          SRCCOPY);             // Copies the source rectangle directly to the destination rectangle.
 // Substituted Canvas.CopyRect with BitBlt for speed up
 // GaugeTempBMP.Canvas.CopyRect(GaugeRect, BasicBMP.Canvas, GaugeRect);

 // Draw spectrum image to GaugeTempBMP
   for j := 1 to HBlockCount do
   begin
      if Bands[j-1] > VLimit then
         Bands[j-1] := VLimit;

      if Bands[j-1] > 0 then
      begin
     // Copy partial image of DisplayBar to GaugeTempBMP
        BarRect.Left := 0;
        BarRect.Right := BlockWidth;
        BarRect.Top := VLimit - Bands[j-1];
        if BarRect.Top < 0 then
           BarRect.Top := 0;
        BarRect.Bottom := DisplayBar.Height;

        tmpRect.Left := (BlockWidth + HBlockGap) * (j - 1) + 2;
        tmpRect.Right := tmpRect.Left + BlockWidth;
        tmpRect.Top := BarRect.Top;
        tmpRect.Bottom := BarRect.Bottom;

        BitBlt(GaugeTempBMP.Canvas.Handle,
               tmpRect.Left,
               tmpRect.Top,
               BlockWidth,
               tmpRect.Bottom - tmpRect.Top + 1,
               DisplayBar.Canvas.Handle,
               BarRect.Left,
               BarRect.Top,
               SRCCOPY);
      end;

      if Bands[j-1] >= trunc(PeakValue[j]) then
      begin
         PeakValue[j] := Bands[j-1] + 0.01;  // 0.01 : to compensate round off
         PassedCounter[j] := 0;
      end else if Bands[j-1] < trunc(PeakValue[j]) then
      begin
         if trunc(PeakValue[j]) > 0 then
         begin
            with GaugeTempBMP.Canvas do
            begin
            // Draw peak line
               Pen.Color := TColor(RGB(192, 192, 192));   // color for peak line
               MoveTo((BlockWidth + HBlockGap) * (j - 1) + 2, VLimit - trunc(PeakValue[j]));
               LineTo((BlockWidth + HBlockGap) * (j - 1) + 2 + BlockWidth, VLimit - trunc(PeakValue[j]));
            end;

      // Followings are to show simillar spectrum image to WINAMP's
      //  - Put delay time before lowering peak line
      //  - Accerate lowering speed according to the time elapsed
            if PassedCounter[j] >= 8 then
                PeakValue[j] := PeakValue[j] - 0.3 * (PassedCounter[j] - 8);

            if PeakValue[j] < 0 then
               PeakValue[j] := 0
            else
               inc(PassedCounter[j]);
         end;
      end;

   end;

 // Copy GaugeTempBMP's image to GaugePaintBox
   BitBlt(GaugePaintBox.Canvas.Handle,
          GaugeRect.Left,
          GaugeRect.Top,
          GaugeTempBMP.Width,
          GaugeTempBMP.Height,
          GaugeTempBMP.Canvas.Handle,
          GaugeRect.Left,
          GaugeRect.Top,
          SRCCOPY);
end;

// Show background image during non-playing period
procedure TMainForm.ShowBackground;
begin
   BitBlt(GaugePaintBox.Canvas.Handle,
          GaugeRect.Left,
          GaugeRect.Top,
          BasicBMP.Width,
          BasicBMP.Height,
          BasicBMP.Canvas.Handle,
          GaugeRect.Left,
          GaugeRect.Top,
          SRCCOPY);
end;


procedure TMainForm.btnPluginSetupClick(Sender: TObject);
begin
   BassPlayer1.ShowPluginConfigForm;
end;

procedure TMainForm.EchoSliderChange(Sender: TObject);
begin
   BassPlayer1.EchoLevel := EchoSlider.Value;
end;

procedure TMainForm.ReverbSliderChange(Sender: TObject);
begin
   BassPlayer1.ReverbLevel := ReverbSlider.Value;
end;

procedure TMainForm.btnShowVisCtrlFormClick(Sender: TObject);
begin
   VisControlForm.Show;
end;

end.
