// Main unit of VisOut.dpr
// The form of this unit is not visible at run time.
// It only takes charge of driving Winamp visualization plug-in.
//    written by Silhwan Hyun  hsh@chollian.net

unit VisOutIntf;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ioplug, PluginCtrl, fisSharedMemory;

type
  TdumbForm = class(TForm)
    fisSharedMemory1: TfisSharedMemory;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
  private
    { Private declarations }
    procedure RenderToVis;
    procedure EndProgram;
  public
    { Public declarations }
    procedure HandleMessages(var Msg: TMsg; var Handled: Boolean);
  end;

var
  dumbForm: TdumbForm;

implementation

{$R *.DFM}

const
   DataReady = 10;
   QuitProg = 20;
var
   ShareMemPointer : pointer;
   DataReadyMsg : hwnd;

   PluginName : string;
   VisModNum : integer;
   Vismod : TVismod;
   NumVismod : integer;


procedure TdumbForm.RenderToVis;
var
   p1 : PDWORD;
   p2 : PBYTE;
   i : integer;
begin
   p1 := ShareMemPointer;

   Vismod[VisModNum]^.sRate := p1^;
   inc(p1, 1);
   Vismod[VisModNum]^.nCh := p1^;
   inc(p1, 1);

   p2 := pointer(p1);
   for i := 1 To 576 do
   begin
     Vismod[VisModNum]^.SpectrumData[1, i] := p2^;
     inc(p2, 1);
   end;
   for i := 1 To 576 do
   begin
     Vismod[VisModNum]^.SpectrumData[2, i] := p2^;
     inc(p2, 1);
   end;

   for i := 1 to 576 do
   begin
     Vismod[VisModNum]^.Waveformdata[1, i] := p2^;
     inc(p2, 1);
   end;
   for i := 1 To 576 do
   begin
     Vismod[VisModNum]^.Waveformdata[2, i] := p2^;
     inc(p2, 1);
   end;

   if Vismod[VisModNum].Render(Vismod[VisModNum]) = 1 then
      Vismod[VisModNum].quit(Vismod[VisModNum]);
end;

procedure TdumbForm.EndProgram;
begin
   if NumVismod > 0 then
      UnloadVisModule;
   NumVismod := 0;
   Close;
end;

procedure TdumbForm.HandleMessages(var Msg: TMsg; var Handled: Boolean);
begin
   if (Msg.Message = DataReadyMsg) then
   begin
      case Msg.wParam of
         DataReady : RenderToVis;
         QuitProg : EndProgram;
      end;
      Handled := True;
   end;
end;

procedure TdumbForm.FormCreate(Sender: TObject);
var
   Visheader : PWinampVisHeader;
   Msg : pchar;
begin
   if not fisSharedMemory1.OpenMemory then
   begin
      Msg := 'Shared memory is not available' + chr(10) +
             '(This program is designed to be called by BASSPlay.exe)';
      MessageBox(Application.Handle, Msg, 'Error', MB_OK + MB_ICONERROR);
      Application.Terminate;
      exit;
   end;

   DataReadyMsg := fisSharedMemory1.SendingDataMsg;
   ShareMemPointer := fisSharedMemory1.MapMemory;
   Application.OnMessage := HandleMessages;

   PluginName := ParamStr(1);
   VisModNum := strToint(ParamStr(2));
   LoadVisModule(PluginName, Visheader, Vismod, NumVismod);
   if (NumVismod > 0) and (VisModNum < NumVismod) and (VisModNum >= 0) then
      StartVisModule(VisModNum)
   else begin
      if NumVismod = 0 then
         Msg := 'Specified plug-in is not available'
      else
         Msg := 'Invalid module number';
      MessageBox(Application.Handle, Msg, 'Error', MB_OK + MB_ICONERROR);
      Application.Terminate;
      exit;
   end;
end;

procedure TdumbForm.FormDestroy(Sender: TObject);
begin
   if ShareMemPointer <> nil then
     fisSharedMemory1.UnMapMemory(ShareMemPointer);
   fisSharedMemory1.CloseMemory;
end;

end.
