unit VisPluginCtrl;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Grids, StdCtrls;

type
  TVisControlForm = class(TForm)
    Label1: TLabel;
    PluginList: TListBox;
    btnConfigure: TButton;
    Label3: TLabel;
    btnStart: TButton;
    Descript: TEdit;
    Label2: TLabel;
    btnStop: TButton;
    VisModules: TStringGrid;
    Label4: TLabel;
    Label5: TLabel;
    Label6: TLabel;
    procedure FormShow(Sender: TObject);
    procedure PluginListClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure btnConfigureClick(Sender: TObject);
    procedure btnStartClick(Sender: TObject);
    procedure btnStopClick(Sender: TObject);
  private
    { Private declarations }
    procedure GetPluginInfo;
  public
    { Public declarations }
  end;

var
  VisControlForm: TVisControlForm;

implementation

{$R *.DFM}

uses
   BassTest, ioPlug, PluginCtrl;

var
   Vismod : TVismod;
   IndexNum,
   NumVismod : integer;

procedure TVisControlForm.GetPluginInfo;
var
   PluginName : string;
   Visheader : PWinampVisHeader;
   i : integer;
begin
   IndexNum := PluginList.ItemIndex;
   PluginName := GetProgDir + 'Plugins\' + PluginList.Items[IndexNum];
   LoadVisModule(PluginName, Visheader, Vismod, NumVismod);
   if NumVismod = 0 then
      exit;

   Descript.Text := string(Visheader^.description);

   VisModules.RowCount := NumVismod;
   for i := 0 to (NumVismod - 1) do
      VisModules.Cells[0, i] := Vismod[i].description;
end;

procedure TVisControlForm.FormShow(Sender: TObject);
var
   SearchRec : TSearchRec;
begin
   IndexNum := -1;
   NumVismod := 0;
   PluginList.Clear;

   if FindFirst(GetProgDir + 'Plugins\vis_*.dll', 0, SearchRec) = 0 then
   begin
      PluginList.Items.Add(SearchRec.Name);
      while FindNext(SearchRec) = 0 do
         PluginList.Items.Add(SearchRec.Name);
   end;
   FindClose(SearchRec);

   if PluginList.Items.Count > 0 then
   begin
      PluginList.ItemIndex := 0;
      GetPluginInfo;
   end else
   begin
      Descript.Text := '';
      VisModules.RowCount := 0;
   end;
end;

procedure TVisControlForm.PluginListClick(Sender: TObject);
begin
   if PluginList.ItemIndex = IndexNum then
      exit;

   UnloadVisModule;
   GetPluginInfo;
end;

procedure TVisControlForm.FormClose(Sender: TObject; var Action: TCloseAction);
begin
   UnloadVisModule;
end;

procedure TVisControlForm.btnConfigureClick(Sender: TObject);
var
   NewSelection : TGridRect;
begin
   if NumVismod = 0 then
      exit;

   NewSelection := VisModules.Selection;
   if NewSelection.Top > (NumVismod - 1) then
      exit;

   Vismod[NewSelection.Top].Config(Vismod[NewSelection.Top]);
end;

procedure TVisControlForm.btnStartClick(Sender: TObject);
begin
   if not FileExists(GetProgDir + 'VisOut.exe') then
   begin
      MessageBox(Application.Handle, 'Visualization program(VisOut.exe) is not present',
                 'Error', MB_OK	+ MB_ICONERROR);
      exit;
   end;
   
   BassPlayer1.RunVisoutProgram(GetProgDir + 'Plugins\' + PluginList.Items[IndexNum],
                                VisModules.Selection.Top);
end;

procedure TVisControlForm.btnStopClick(Sender: TObject);
begin
   BassPlayer1.QuitVisoutProgram;
end;

end.
